import json
import textwrap
from contextlib import contextmanager
from pathlib import Path

from graphql.type.definition import (
    GraphQLEnumType,
    GraphQLNonNull,
    GraphQLScalarType,
)

from .helpers import deep_get

gql_folder = Path(__file__).parent / "gql"


def first_item(dct):
    return list(dct.items())[0]


def node_is_terminal(obj):
    if isinstance(obj, (GraphQLNonNull)):
        return node_is_terminal(obj.of_type)
    elif isinstance(obj, (GraphQLScalarType, GraphQLEnumType)):
        return True
    else:
        return False


def generate_default_node(obj):
    if isinstance(obj, GraphQLScalarType):
        return ""

    keys = "\n".join(
        [key for key, field in obj.fields.items() if node_is_terminal(field.type)]
    )

    return "{\n" + keys + "\n}"


def format_node_from_dicts(node):
    if node:
        fields = "\n".join(
            [f"{key} {format_node_from_dicts(value)}" for key, value in node.items()]
        )
        return "{\n" + fields + "\n}\n"
    else:
        return ""


# def generate_list_node(node_type, support_templates):
#     fields = get_list_field_config(node_type, support_templates)
#     node = {}

#     for field in fields:
#         parts = field.split(".")
#         root = node
#         stem, leaf = parts[:-1], parts[-1]
#         for part in stem:
#             if part not in root:
#                 root[part] = {}
#             root = root[part]
#         root[leaf] = None

#     return format_node_from_dicts(node)


# def get_list_field_config(node_type, support_templates):
#     list_fields = support_templates["ListFields"]
#     default_fields = ["id", "name", "label", "description"]
#     return list_fields.get(node_type, default_fields)


# def filter_list_fields(result, node_type, support_templates):
#     fields = get_list_field_config(node_type, support_templates)

#     new_nodes = []
#     for edge in result["edges"]:
#         node = edge["node"]
#         row = {}
#         for field in fields:
#             field_parts = field.split(".")
#             if field_parts[0] in node:
#                 collapsed = node[field_parts[0]]
#                 row[field] = deep_get(collapsed, field_parts[1:])

#         new_nodes.append(row)

#     return new_nodes


class IndentWriter:
    """
    A simple utility to help with witing indented structures
    """

    def __init__(self):
        self.buffer = []
        self.level = 0

    @contextmanager
    def indent(self):
        self.increment()
        yield
        self.decrement()

    @contextmanager
    def space(self):
        self("")
        yield
        self("")

    def increment(self):
        self.level += 1

    def decrement(self):
        self.level -= 1

    def __call__(self, content):
        if isinstance(content, str):
            self([content])
        else:
            self.buffer.extend(["  " * self.level + line for line in content])

    def result(self):
        return "\n".join(self.buffer)


def write_page_info(wrt):
    wrt("pageInfo {")
    with wrt.indent():
        wrt("hasNextPage")
        wrt("hasPreviousPage")
        wrt("startCursor")
        wrt("endCursor")

    wrt("}")


def format_graphql(call_type, args, field_name, node_path, node_gql):
    """
    Formats a graphQL statement based on its parts
    """

    wrt = IndentWriter()

    add_page_info = call_type == "query"
    # Strips off the edge braces and clears and empty lines
    node_gql_body = list(
        filter(
            lambda line: line,
            textwrap.dedent(node_gql.strip().strip("{").strip("}")).split("\n"),
        )
    )

    wrt(f"{call_type}(")
    with wrt.indent():
        wrt([(f"${name}: {type}") for name, type in args])
    wrt(") {")

    with wrt.space():
        with wrt.indent():
            wrt(f"{field_name}(")
            with wrt.indent():
                wrt([(f"{name}: ${name}") for name, _ in args])
            wrt(") {")

            with wrt.indent():
                with wrt.space():
                    for part in node_path[:-1]:
                        wrt(part + " {")
                        wrt.increment()
                    if node_gql_body:
                        wrt(node_path[-1] + " {")
                        with wrt.space():
                            with wrt.indent():
                                wrt(node_gql_body)
                        wrt("}")
                    else:
                        wrt(node_path[-1])

                    for part in node_path[1:]:
                        wrt.decrement()
                        wrt("}")

                    if add_page_info:
                        write_page_info(wrt)
            wrt("}")

    wrt("}")

    result = wrt.result()
    return result


class GQLGenerator:
    def __init__(self, schema, templates):
        self.templates = templates
        self.schema = schema

    def generate_node(self, node_type):
        gql_templates = self.templates["GQL"]
        node_type_name = str(node_type)
        if node_type_name in gql_templates:
            return gql_templates[node_type_name]

        return generate_default_node(node_type)

    def get_list_field_config(self, node_type):
        list_fields = self.templates["ListFields"]
        default_fields = ["id", "name", "label", "description"]
        return list_fields.get(str(node_type), default_fields)

    def generate_list_node(self, node_type):
        fields = self.get_list_field_config(node_type)
        node = {}

        for field in fields:
            parts = field.split(".")
            root = node
            stem, leaf = parts[:-1], parts[-1]
            for part in stem:
                if part not in root:
                    root[part] = {}
                root = root[part]
            root[leaf] = None

        return format_node_from_dicts(node)

    def filter_list_fields(self, field_name, result):
        root = self.schema.query_type
        root_field = root.fields[field_name]
        field_type = root_field.type

        root_return_type = (
            field_type.fields["edges"].type.of_type.of_type.fields["node"].type
        )
        fields = self.get_list_field_config(root_return_type)
        new_nodes = []
        for node in result:
            row = {}
            for field in fields:
                field_parts = field.split(".")
                if field_parts[0] in node:
                    collapsed = node[field_parts[0]]
                    row[field] = deep_get(collapsed, field_parts[1:])

            new_nodes.append(row)

        return new_nodes

    def generate_endpoint(self, call_type, field_name):

        if call_type == "mutation":
            root = self.schema.mutation_type
        elif call_type == "query":
            root = self.schema.query_type
        else:
            raise Exception("Unknown call type")

        root_field = root.fields[field_name]

        args = [(arg_name, str(arg.type)) for arg_name, arg in root_field.args.items()]

        field_type = root_field.type

        if call_type == "mutation":

            root_return_key, root_return_field = first_item(field_type.fields)
            root_return_type = root_return_field.type
            return_node = self.generate_node(root_return_type)
            return format_graphql(
                "mutation", args, field_name, [root_return_key], return_node
            )

        elif call_type == "query":
            # Basically this gets us edges.node while going through
            # all the NonNull and List Shit
            root_return_type = (
                field_type.fields["edges"].type.of_type.of_type.fields["node"].type
            )
            return_node = self.generate_node(root_return_type)
            return format_graphql(
                "query", args, field_name, ["edges", "node"], return_node
            )

    def generate_table_endpoint(self, call_type, field_name):
        root = self.schema.query_type
        root_field = root.fields[field_name]

        args = [(arg_name, str(arg.type)) for arg_name, arg in root_field.args.items()]

        field_type = root_field.type
        root_return_type = (
            field_type.fields["edges"].type.of_type.of_type.fields["node"].type
        )
        return_node = self.generate_list_node(root_return_type.name)

        return format_graphql("query", args, field_name, ["edges", "node"], return_node)



import time
import sys


def deep_get(obj, path):
    if not path:
        return obj
    elif isinstance(obj, dict):
        return deep_get(obj[path[0]], path[1:])
    elif isinstance(obj, list):
        return [deep_get(child, path) for child in obj]
    else:
        return obj

def uppercase_dict(d):
    """
    Returns a dict with all teh keys, and any child keys uppercased.

    This is used for case insensitive lookups.
    """
    return {
        k.upper(): (uppercase_dict(v) if isinstance(v, dict) else v)
        for k, v in d.items()
    }


"""
This module exposes a local runtime environment for RelyComply risk rating functions that
emulates the environment of the RelyComply cloud.
"""

import pprint
from dataclasses import dataclass
from datetime import datetime, timezone
import json

import tabulate
from relycomply_client.rc import rc


def print_header(header):
    print()
    print(header)
    print("=" * len(header))


@dataclass
class Score:
    """
    The return value of a a score_customer function.
    """

    name: str  # The name of the score
    label: str  # A human readable label for the score
    value: float  # The numeric value of the score
    maximum: float  # The maximum value the score can take
    minimum: float = 0  # The minimum value the score can take
    category: str = None  # The category of the score (high,medium,low)
    components: list["Score"] = None  # A list of sub scores that make up this score


class CustomerFacade:
    """
    This is a convenience wrapper around the customer API packet.
    """

    def __init__(self, customer_packet):
        self.__dict__.update(customer_packet)

    def customer_data(self, key, multiple_values=False):
        """
        Given a customer data key return the associated value.

        If multiple_values is True, return a list of all values associated with the key.
        """
        results = [datum["value"] for datum in self.data if datum["key"] == key]

        if not multiple_values:
            if len(results) > 1:
                raise ValueError(f"Multiple values found for key {key}")
            elif len(results) == 0:
                raise ValueError(f"No values found for key {key}")
            return results[0]

        return results

    def indicator(self, key):
        """
        Given a customer indicator key return whether the customer has the indicator.
        """
        indicators = self.indicators

        now = datetime.now(timezone.utc)
        for indicator in indicators:
            name = indicator["indicator"]["name"]
            valid_from = datetime.fromisoformat(indicator["validFrom"])
            valid_until = (
                None
                if indicator["validUntil"] is None
                else datetime.fromisoformat(indicator["validUntil"])
            )

            if (
                name == key
                and valid_from < now
                and (valid_until is None or valid_until > now)
            ):
                return True

        return False


class GraphQLLookupTableQuery:
    """
    This creates a callable that uses the graphQL API to query lookup tables associated with a risk rating.
    This emulates the same behaviour as the RelyComply risk rating engine.

    it is exposed to the score customers function as a lookup function with the following signature:

    table_name:         string of the lookup table name
    value_column: the   column to return value or values from. This is what you are looking up.
    multiple_values:    If true, return a list of values. If false, return a single value.
    **filters:          key value pairs of filters to apply to the lookup table. The key here corresponds
                        to the column name you wish to filter. The value can either be a scalar or a list of scalars.
                        If a single value is given it will filter rows where only that value is matched. If a list
                        is given it will filter rows where any of teh given values are present.

    Example usage in a score_customer function:

    ```python
    def score_customer(parameters, customer, lookup, Score):
        ...
        jurisdiction_scores = lookup(
            "jurisdiction_scores",
            client_type=customer.customer_data("client_type"),
            country=[customer.nationality, customer.residence]
            value_column="score",
            multiple_values=True
        )

        jurisdiction_score = max(jurisdiction_scores)
        ...

    ```
    """

    def __init__(self, risk_rating):
        self.risk_rating = risk_rating

    def __call__(self, table_name, *, value_column, multiple_values=False, **filters):
        print_header(f"Lookup on {table_name}")

        list_filters = [
            dict(
                column=key,
                values=[str(value) for value in values]
                if isinstance(values, list)
                else [str(values)],
            )
            for key, values in filters.items()
        ]

        matching_lookups = [
            lookup
            for lookup in self.risk_rating["lookups"]
            if lookup["table"]["name"] == table_name
        ]
        if not matching_lookups:
            raise ValueError(
                f"Lookup table {table_name} not found in the lookups associated with this risk rating config"
            )
        if len(matching_lookups) > 1:
            raise ValueError(
                f"Multiple lookup tables found with name {table_name} in the lookups associated with this risk rating config"
            )
        version = matching_lookups[0]["versionNumber"]

        variables = dict(
            table=table_name,
            version=str(version),
            filters=list_filters,
            valueColumn=value_column,
        )

        pprint.pprint(variables)
        print()

        response = rc.gql.graphql(
            graphql_query="""
                query (
                    $table: LookupTableReference!, 
                    $version: String!, 
                    $filters: [LookupTableQueryFilterInput]!, 
                    $valueColumn: String!
                ) {
                    queryLookupTable(table: $table, version: $version, filters: $filters, valueColumn: $valueColumn) {
                        count
                        rows
                        schema {
                            columns {
                                name
                                dtype
                            }
                        }
                        valuesString
                        valuesNumber
                    }
                }
            """,
            variables=variables,
        )

        results = response["queryLookupTable"]

        print(
            tabulate.tabulate(
                [json.loads(row).values() for row in results["rows"]],
                headers=[column["name"] for column in results["schema"]["columns"]],
            )
        )

        if results["count"] == 0:
            return ValueError("No values found")
        if not multiple_values and results["count"] > 1:
            raise ValueError("Multiple values found")

        values = results["valuesNumber"] or results["valuesString"]

        if multiple_values:
            return values
        else:
            return values[0]


def print_score(score):
    def map_score(s):
        return [s.name, s.label, s.value, s.minimum, s.maximum, s.category]

    print(
        tabulate.tabulate(
            [map_score(score), *map(map_score, score.components)],
            headers=["Name", "Label", "Value", "Minimum", "Maximum", "Category"],
        )
    )


def run_risk_rating(scoring_function, customer_identifier, risk_rating_config_id):
    """
    Given a scoring function and a customer identifier and risk rating config id, run the scoring function.

    This will build an execution environment that works in the same way as the risk rating engine
    on the RelyComply cloud. It allows for simple local debugging and testing of risk rating configurations.

    Usage

    ```python
    from relycomply_client.risk_rating_runtime import run_risk_rating
    from my_risK_rating_module import score_customer

    run_risk_rating(scoring_function, "<customer_identifier>", "<risk_rating_id_or_name>")

    ```
    """
    print_header("Customer")
    customer_packet = rc.gql.customers(identifier=str(customer_identifier), _only=True)
    pprint.pp(customer_packet)

    print_header("Risk rating config")
    risk_rating = rc.gql.riskRatingConfigs(id=risk_rating_config_id, _only=True)
    pprint.pp(risk_rating)

    print_header("Parameters")
    parameters = {
        parameter["key"]: parameter["value"] for parameter in risk_rating["parameters"]
    }
    pprint.pp(parameters)

    lookup = GraphQLLookupTableQuery(risk_rating)
    customer = CustomerFacade(customer_packet)
    score = scoring_function(
        parameters=parameters, customer=customer, lookup=lookup, Score=Score
    )

    print_header("Final scores")
    print_score(score)

    return score



import logging
import os
from contextlib import contextmanager
from pathlib import Path
from typing import Optional

import toml
import typer
from tabulate import tabulate

from relycomply_client.cli import RelyComplyCLI
from relycomply_client.exceptions import (
    RelyComplyClientException,
    RelyComplyCliException,
)
from relycomply_client.relycomply_gql_client import RelyComplyGQLClient


@contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)


class RelySync:
    def __init__(
        self,
        location,
        recursive,
        interactive,
        cli: RelyComplyCLI = None,
        break_on_error=True,
        create_only=False,
        debug=False,
    ):
        if debug:
            logging.root.setLevel(logging.DEBUG)
            logging.root.addHandler(logging.StreamHandler())

        self.location = location or Path(".")
        self.recursive = recursive
        self.interactive = interactive
        self.break_on_error = break_on_error
        self.create_only = create_only

        self.cli: RelyComplyCLI = cli or RelyComplyCLI()
        self.gql: RelyComplyGQLClient = self.cli.gql

        self.sync_order: list[str] = self.gql.templates["TypeRelationships"][
            "SyncOrder"
        ]
        self.sync_order_upper: list[str] = [item.upper() for item in self.sync_order]

        self.implementations: list[str] = self.gql.templates["TypeRelationships"][
            "Implementations"
        ]

    def run_cli(self):
        self.locate_files()
        self.display_locations()

        if not self.with_metadata:
            return

        self.schedule_actions()

        self.display_schedule()

        if self.interactive and not typer.prompt(
            "Are you sure you would like to continue (yes/no)?", type=bool
        ):
            return

        self.execute_schedule()

    def locate_files(self):
        # These are the supported types

        if isinstance(self.location, list):
            toml_files = [Path(raw_path) for raw_path in self.location]
        elif self.location.is_dir():
            glob_pattern = "**/*.toml" if self.recursive else "*.toml"
            toml_files = list(self.location.glob(glob_pattern))
        elif self.location.is_file():
            toml_files = [self.location]

        self.no_metadata = set()
        self.incorrect_metadata = []
        self.with_metadata = []
        for path in toml_files:
            with open(path) as f:
                lines = f.readlines()

                # Clear empty lines
                lines = [line.strip() for line in lines if line.strip()]

                # Strip preceeding normal comments
                lines = [
                    line
                    for line in lines
                    if not (line[0] == "#" and not line.startswith("#%"))
                ]

                # Has no fore-matter
                if not lines or not lines[0].startswith("#%"):
                    self.no_metadata.add(path)

                fore_lines = []
                for line in lines:
                    if line.startswith("#%"):
                        fore_lines.append(line[2:].strip())
                    else:
                        break

                metadata_str = "\n".join(fore_lines)

                try:
                    metadata = toml.loads(metadata_str)
                    if not metadata:
                        self.no_metadata.add(path)
                    elif "type" not in metadata:
                        self.incorrect_metadata.append((path, "No type information"))
                    elif (
                        "type" in metadata
                        and metadata["type"].upper() not in self.sync_order_upper
                    ):
                        self.incorrect_metadata.append(
                            (
                                path,
                                f"Type '{metadata['type']}' is not support for syncing",
                            )
                        )
                    else:
                        self.with_metadata.append((path, metadata))

                except Exception as e:
                    self.incorrect_metadata.append(
                        (path, f"Metadata TOML error: {str(e)}")
                    )

    def display_locations(self):
        if self.no_metadata:
            typer.secho("Found the following files with no metadata")
            for path in self.no_metadata:
                typer.secho("  - " + str(path))
            typer.echo()

        if self.incorrect_metadata:
            typer.secho(
                "Found the following files with incorrect metadata", fg=typer.colors.RED
            )
            for path, error in self.incorrect_metadata:
                typer.secho(f"  - {path} [{error}]", fg=typer.colors.RED)
            typer.echo()

        if self.with_metadata:
            typer.secho(
                "Found the following files with type information", fg=typer.colors.GREEN
            )
            for path, metadata in sorted(self.with_metadata):
                typer.secho(
                    f"  - {path} [type = {metadata['type']}]", fg=typer.colors.GREEN
                )
            typer.echo()

    def get_rank(self, path, metadata):
        return self.sync_order_upper.index(metadata["type"].upper()), path

    def schedule_actions(self):
        self.with_metadata.sort(key=lambda t: self.get_rank(*t))

        implements_map = {
            implementor: implemented
            for implemented, implementors in self.implementations.items()
            for implementor in implementors
        }

        for path, metadata in self.with_metadata:
            metadata["implements"] = implements_map.get(
                metadata["type"], metadata["type"]
            )

        self.schedule = [
            dict(path=path, metadata=metadata, variables=toml.load(path))
            for path, metadata in self.with_metadata
        ]

        for item in self.schedule:
            create = False

            try:
                if "name" in item["variables"]:
                    key = "name"
                elif "identifier" in item["variables"]:
                    key = "identifier"
                else:
                    key = None

                item["metadata"]["key"] = key

                if key:
                    existing = self.cli.execute(
                        item["metadata"]["implements"],
                        "retrieve",
                        {key: item["variables"][key]},
                    )
                    item["action"] = "update"
                    item["existing"] = existing
                else:
                    create = True
            except RelyComplyCliException:
                create = True

            if create:
                item["action"] = "create"
                item["existing"] = None

    def display_schedule(self):
        to_create = [item for item in self.schedule if item["action"] == "create"]
        to_update = [item for item in self.schedule if item["action"] == "update"]

        def display_schedule_table(items, message):
            if items:
                typer.secho(message)
                typer.echo()
                typer.echo(
                    tabulate(
                        [
                            [
                                item["metadata"]["type"],
                                item["variables"].get("name"),
                                item["path"],
                            ]
                            for item in items
                        ],
                        headers=["Type", "name", "path"],
                    )
                )
                typer.echo()

        display_schedule_table(to_create, "The following items will be created:")
        if not self.create_only:
            display_schedule_table(to_update, "The following items may be updated:")

    def execute_schedule(self):
        for item in self.schedule:
            try:
                with pushd(str(item["path"].parent)):
                    if item["action"] == "create":
                        typer.echo(
                            f'rely {item["metadata"]["type"]} {item["action"]} {item["path"]}'
                        )
                        self.cli.execute(
                            item["metadata"]["type"], item["action"], item["variables"]
                        )
                    elif item["action"] == "update" and not self.create_only:
                        typer.echo(
                            f'rely {item["metadata"]["type"]} {item["action"]} --id={item["variables"][item["metadata"]["key"]]} {item["path"]}'
                        )

                        self.cli.execute(
                            item["metadata"]["type"],
                            item["action"],
                            {
                                "id": item["variables"][item["metadata"]["key"]],
                                **item["variables"],
                            },
                        )
            except RelyComplyClientException as e:
                typer.secho(str(e), fg=typer.colors.RED)
                if self.break_on_error:
                    break
            except FileNotFoundError as e:
                typer.secho(f"File not found: {e.filename}", fg=typer.colors.RED)
                typer.secho(f"Referenced in file: {item['path']}", fg=typer.colors.RED)
                if self.break_on_error:
                    break


def sync(
    location: Optional[Path] = typer.Argument(
        Path("."), help="The location to look for configuration files"
    ),
    recursive: bool = typer.Option(
        False,
        help="If true will search for files recursively from the given location directory",
    ),
    interactive: bool = typer.Option(
        True, help="If true the will ask for confirmation before taking any action"
    ),
    break_on_error: bool = typer.Option(
        True,
        help="If true the sync will stop on the first error. If false it will attempt to continue.",
    ),
    create_only: bool = typer.Option(False, help="If true will only create new items"),
    debug: bool = typer.Option(False, help="Enable debug logging"),
):
    """
    This will attempt to synchronise the specified configuration files with the relycomply server.
    It will (semi-)intelligently determine if the file should be created or updated based on
    the name/identifier in the file.

    The file should have a fore-matter block at the top of the file that is TOML formatted.
    This fore-matter should specify the type of the object in the format:

    #% type="Type"

    for example:

    #% type="Product"
    name = "bank_account"
    label = "Bank Account"

    """
    rely_sync = RelySync(
        location=location,
        recursive=recursive,
        interactive=interactive,
        break_on_error=break_on_error,
        debug=debug,
        create_only=create_only,
    )
    rely_sync.run_cli()


def main():
    typer.run(sync)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Optional

import typer
from watchfiles import Change, DefaultFilter, watch

from relycomply_client.cli import RelyComplyCLI
from relycomply_client.sync import RelySync


class TomlFilter(DefaultFilter):
    allowed_extensions = ".toml"

    def __call__(self, change: Change, path: str) -> bool:
        return super().__call__(change, path) and path.endswith(self.allowed_extensions)


def watch_sync(
    location: Optional[Path] = typer.Argument(
        Path("."), help="The location to look for configuration files"
    ),
    recursive: bool = typer.Option(
        False,
        help="If true will search for files recursively from the given location directory",
    ),
    interactive: bool = typer.Option(
        False, help="If true the will ask for confirmation before taking any action"
    ),
    break_on_error: bool = typer.Option(
        True,
        help="If true the sync will stop on the first error. If false it will attempt to continue.",
    ),
):
    """
    Watch for changes in configuration files and sync them with rely-sync
    """
    cli = RelyComplyCLI()

    abs_location = location.absolute()
    wait_message = f"Waiting for changes at path: {abs_location} {'(recursive)' if recursive else ''}\n"

    print(wait_message)

    for changes in watch(
        location, recursive=recursive, debounce=300, watch_filter=TomlFilter()
    ):
        print("Detected changes in:")
        for change in changes:
            print(f"  - {change[1]}")
        print()
        changed_locations = [change[1] for change in changes]
        rely_sync = RelySync(
            changed_locations,
            recursive,
            interactive,
            cli=cli,
            break_on_error=break_on_error,
        )
        rely_sync.run_cli()
        print()
        print(wait_message)


def main():
    typer.run(watch_sync)


if __name__ == "__main__":
    main()
